package org.dasein.cloud.vsphere;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;

import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.services.dc.DataCenter;
import org.dasein.cloud.services.dc.DataCenterServices;
import org.dasein.cloud.services.dc.Region;

import com.vmware.vim25.InvalidProperty;
import com.vmware.vim25.RuntimeFault;
import com.vmware.vim25.mo.Datacenter;
import com.vmware.vim25.mo.Folder;
import com.vmware.vim25.mo.InventoryNavigator;
import com.vmware.vim25.mo.ManagedEntity;
import com.vmware.vim25.mo.ServiceInstance;

public class Dc implements DataCenterServices {

    private PrivateCloud provider;
    
    Dc(PrivateCloud cloud) {
        provider = cloud;
    }
    
    @Override
    public DataCenter getDataCenter(String dcId) throws InternalException, CloudException {
        ServiceInstance service = provider.getServiceInstance();
        
        try {
            return toDataCenter(getVmwareDatacenter(service, dcId));
        }
        finally {
            if( service != null ) {
                service.getServerConnection().logout();
            }
        }
    }

    @Override
    public String getProviderTermForDataCenter(Locale locale) {
        return "data center";
    }

    @Override
    public String getProviderTermForRegion(Locale locale) {
        return "region";
    }

    @Override
    public Region getRegion(String regionId) throws InternalException, CloudException {
        for( Region region : listRegions() ) {
            if( region.getProviderRegionId().equals(regionId) ) {
                return region;
            }
        }
        return null;
    }

    
    Datacenter getVmwareDatacenter(ServiceInstance service, String dcId) throws CloudException, InternalException {
        Folder rootFolder = service.getRootFolder();
        
        try {
            return (Datacenter)(new InventoryNavigator(rootFolder).searchManagedEntity("Datacenter", dcId));
        }
        catch( InvalidProperty e ) {
            throw new InternalException("Invalid DC property: " + e.getMessage());
        }
        catch( RuntimeFault e ) {
            throw new InternalException("Error talking to the cluster: " + e.getMessage());
        }
        catch( RemoteException e ) {
            throw new CloudException("Error in processing the request in the cluster: " + e.getMessage());
        }
    }
    
    @Override
    public Collection<DataCenter> listDataCenters(String regionId) throws InternalException, CloudException {
        ArrayList<DataCenter> dataCenters = new ArrayList<DataCenter>();
        ServiceInstance instance = provider.getServiceInstance();
            
        try {
            Folder rootFolder = instance.getRootFolder();
            ManagedEntity[] mes;
            
            try {
                mes = new InventoryNavigator(rootFolder).searchManagedEntities("Datacenter");
            }
            catch( InvalidProperty e ) {
                throw new CloudException("No virtual machine support in cluster: " + e.getMessage());
            }
            catch( RuntimeFault e ) {
                throw new CloudException("Error in processing request to cluster: " + e.getMessage());
            }
            catch( RemoteException e ) {
                throw new CloudException("Error in cluster processing request: " + e.getMessage());
            }
            
            if( mes == null || mes.length < 1 ) {
                return dataCenters;
            }
            for( ManagedEntity entity : mes ) {
                DataCenter dc = toDataCenter((Datacenter)entity);
                
                if( dc != null ) {
                    dataCenters.add(dc);
                }
            }
            return dataCenters;
        }
        finally {
            if( instance != null ) {
                instance.getServerConnection().logout();
            }
        }            
    }

    @Override
    public Collection<Region> listRegions() throws InternalException, CloudException {
        ArrayList<Region> regions = new ArrayList<Region>();
        
        for( String endpoint : provider.getEndpoints() ) {
            Region region = toRegion(provider.getRegionId(endpoint));
            
            if( region != null ) {
                regions.add(region);
            }
        }
        return regions;
    }

    @Override
    public String matchDataCenter(CloudProvider withProvider, String dataCenterId) throws InternalException, CloudException {
        return dataCenterId;
    }

    private DataCenter toDataCenter(Datacenter dc) {
        if( dc != null ) {
            DataCenter dsDc = new DataCenter();
            
            dsDc.setActive(true);
            dsDc.setAvailable(true);
            dsDc.setName(dc.getName());
            dsDc.setProviderDataCenterId(dc.getName());
            dsDc.setRegionId("home");
            return dsDc;
        }
        return null;
    }
    
    private Region toRegion(String regionId) {
        Region region = new Region();
        
        region.setActive(true);
        region.setAvailable(true);
        region.setName(regionId);
        region.setProviderRegionId(regionId);
        return region;        
    }
}
