package org.dasein.cloud.vsphere;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;

import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.services.image.ImageServices;
import org.dasein.cloud.services.image.ServerImage;
import org.dasein.cloud.services.server.Platform;
import org.dasein.cloud.services.server.ServerSize;
import org.dasein.cloud.services.storage.CloudStoreObject;
import org.dasein.util.Jiterator;
import org.dasein.util.JiteratorPopulator;
import org.dasein.util.PopulatorThread;

import com.vmware.vim25.InvalidProperty;
import com.vmware.vim25.RuntimeFault;
import com.vmware.vim25.VirtualMachineConfigInfo;
import com.vmware.vim25.VirtualMachineGuestOsIdentifier;
import com.vmware.vim25.mo.Folder;
import com.vmware.vim25.mo.InventoryNavigator;
import com.vmware.vim25.mo.ManagedEntity;
import com.vmware.vim25.mo.ServiceInstance;
import com.vmware.vim25.mo.VirtualMachine;

public class Template implements ImageServices {

    private PrivateCloud provider;
    
    Template(PrivateCloud cloud) {
        provider = cloud;
    }
    
    @Override
    public String create(String location) throws InternalException, CloudException {
        if( !location.startsWith("instance:") ) {
            throw new CloudException("Must specify an instance for creating a custom template.");
        }

        ServiceInstance service = provider.getServiceInstance();
        String[] parts = location.split(":");
                
        try {
            VirtualMachine vm = ((Vm)provider.getServerServices()).getVirtualMachine(service, parts[1]);
            String name = vm.getName() + "-tpl-" + System.currentTimeMillis();
            
            return toServerImage(((Vm)provider.getServerServices()).clone(service, vm, name, true, false)).getProviderServerImageId();
        }
        finally {
            service.getServerConnection().logout();
        }
    }

    @Override
    public void delete(String templateId) throws InternalException, CloudException {
        // TODO Auto-generated method stub
        
    }

    @Override
    public Collection<ServerImage> find(String executableByAccountNumber, String keyword, Platform platform, String architecture) throws InternalException, CloudException {
        ArrayList<ServerImage> matches = new ArrayList<ServerImage>();
        
        for( ServerImage image : list(executableByAccountNumber) ) {
            if( platform != null ) {
                if( !platform.equals(Platform.UNKNOWN) ) {
                    Platform imgPlatform = image.getPlatform();
            
                    if( !imgPlatform.equals(platform) ) {
                        if( platform.isWindows() ) {
                            continue;
                        }
                        if( platform.equals(Platform.UNIX) && !imgPlatform.isUnix() ) {
                            continue;
                        }
                    }
                }
            }
            if( architecture != null && !architecture.equals(image.getArchitecture()) ) {
                continue;
            }
            if( keyword != null ) {
                keyword = keyword.toLowerCase().trim();
                if( !image.getProviderServerImageId().toLowerCase().contains(keyword) ) {
                    if( !image.getName().toLowerCase().contains(keyword) ) {
                        if( !image.getDescription().toLowerCase().contains(keyword) ) {
                            continue;
                        }
                    }
                }
            }
            matches.add(image);
        }
        return matches;
    }

    @Override
    public ServerImage getImage(String templateId) throws InternalException, CloudException {
        for( ServerImage image : list(null) ) {
            if( image.getProviderServerImageId().equals(templateId) ) {
                return image;
            }
        }
        return null;
    }

    @Override
    public CloudStoreObject getManifest(String location) throws CloudException, InternalException {
        return null;
    }

    @Override
    public String getProviderTermForImage(Locale locale) {
        return "template";
    }

    @Override
    public Collection<String> getShares(String templateId) throws CloudException, InternalException {
        return new ArrayList<String>();
    }

    @Override
    public AsynchronousTask<ServerImage> imageServer(String serverId, String forAccountAccessKey, String location, String withName) throws InternalException, CloudException {
        throw new CloudException("Imaging servers to cloud storage is not supported.");
    }

    @Override
    public boolean isPublic(String templateId) throws InternalException, CloudException {
        return false;
    }

    @Override
    public boolean isSupportsCustomImages() throws InternalException, CloudException {
        return false;
    }

    @Override
    public boolean isSupportsSharing() throws InternalException, CloudException {
        return false;
    }

    @Override
    public Collection<ServerImage> list(String forAccountId) throws InternalException, CloudException {
        PopulatorThread<ServerImage> populator;
        
        populator = new PopulatorThread<ServerImage>(new JiteratorPopulator<ServerImage>() {
            public void populate(Jiterator<ServerImage> iterator) throws CloudException, InternalException {
                populateImageList(iterator);
            }
        });
        populator.populate();
        return populator.getResult();
    }

    private void populateImageList(Jiterator<ServerImage> iterator) throws CloudException, InternalException {
        ServiceInstance instance = provider.getServiceInstance();
        
        try {
            Folder rootFolder = instance.getRootFolder();
            ManagedEntity[] mes;
            
            try {
                mes = new InventoryNavigator(rootFolder).searchManagedEntities("VirtualMachine");
            }
            catch( InvalidProperty e ) {
                throw new CloudException("No virtual machine support in cluster: " + e.getMessage());
            }
            catch( RuntimeFault e ) {
                throw new CloudException("Error in processing request to cluster: " + e.getMessage());
            }
            catch( RemoteException e ) {
                throw new CloudException("Error in cluster processing request: " + e.getMessage());
            }
            
            if( mes != null && mes.length > 0 ) {
                for( ManagedEntity entity : mes ) {
                    VirtualMachine template = (VirtualMachine)entity;

                    if( template != null ) {
                        VirtualMachineConfigInfo cfg = template.getConfig();
                        
                        if( cfg != null && cfg.isTemplate() ) {
                            ServerImage image = toServerImage(template);
                
                            if( image != null ) {
                                iterator.push(image);
                            }
                        }
                    }
                }
            }
        }
        finally {
            if( instance != null ) {
                instance.getServerConnection().logout();
            }
        }
    }
    
    @Override
    public void share(String templateId, String withAccountId, boolean grant) throws InternalException, CloudException {
        throw new OperationNotSupportedException("Sharing not supported.");
    }
    
    private ServerImage toServerImage(VirtualMachine template) throws InternalException, CloudException {
        if( template != null ) {
            VirtualMachineConfigInfo vminfo = template.getConfig();
            VirtualMachineGuestOsIdentifier os = VirtualMachineGuestOsIdentifier.valueOf(vminfo.getGuestId());
            ServerImage image = new ServerImage();
            
            image.setActive(true);
            image.setArchitecture(((Vm)provider.getServerServices()).getArchitecture(os));
            image.setAvailable(true);
            image.setDescription(template.getName());
            image.setName(template.getName());
            image.setOwnerAccountId(provider.getContext().getAccountNumber());
            image.setPersistent(true);
            image.setPlatform(Platform.guess(vminfo.getGuestFullName()));
            image.setProviderServerImageId(vminfo.getUuid());
            image.setPublicImage(false);
            Collection<ServerSize> sizes = provider.getServerServices().getSupportedSizes(image.getArchitecture());

            image.setSize(sizes.toArray(new ServerSize[sizes.size()]));
            image.setType("template");
            return image;
        }
        return null;
    }

}
