package org.dasein.cloud.vsphere;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.Locale;
import java.util.Random;

import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.services.dc.DataCenter;
import org.dasein.cloud.services.image.ServerImage;
import org.dasein.cloud.services.server.Architecture;
import org.dasein.cloud.services.server.Platform;
import org.dasein.cloud.services.server.Server;
import org.dasein.cloud.services.server.ServerPersistence;
import org.dasein.cloud.services.server.ServerServices;
import org.dasein.cloud.services.server.ServerSize;
import org.dasein.cloud.services.server.ServerState;
import org.dasein.cloud.services.server.ServerStatistics;

import com.vmware.vim25.GuestInfo;
import com.vmware.vim25.InvalidProperty;
import com.vmware.vim25.InvalidState;
import com.vmware.vim25.ManagedEntityStatus;
import com.vmware.vim25.RuntimeFault;
import com.vmware.vim25.TaskInProgress;
import com.vmware.vim25.VirtualHardware;
import com.vmware.vim25.VirtualMachineCloneSpec;
import com.vmware.vim25.VirtualMachineConfigInfo;
import com.vmware.vim25.VirtualMachineConfigSpec;
import com.vmware.vim25.VirtualMachineGuestOsIdentifier;
import com.vmware.vim25.VirtualMachinePowerState;
import com.vmware.vim25.VirtualMachineRelocateSpec;
import com.vmware.vim25.VirtualMachineRuntimeInfo;
import com.vmware.vim25.mo.ClusterComputeResource;
import com.vmware.vim25.mo.Datacenter;
import com.vmware.vim25.mo.Folder;
import com.vmware.vim25.mo.HostSystem;
import com.vmware.vim25.mo.InventoryNavigator;
import com.vmware.vim25.mo.ManagedEntity;
import com.vmware.vim25.mo.ResourcePool;
import com.vmware.vim25.mo.ServiceInstance;
import com.vmware.vim25.mo.Task;
import com.vmware.vim25.mo.VirtualMachine;

public class Vm implements ServerServices {

    private PrivateCloud provider;
    
    Vm(PrivateCloud provider) {
        this.provider = provider;
    }
    
    @Override
    public void boot(String serverId) throws InternalException, CloudException {
        ServiceInstance instance = provider.getServiceInstance();
        
        try {
            VirtualMachine vm = getVirtualMachine(instance, serverId);
        
            if( vm != null ) {
                try {
                    Datacenter dc = getVmwareDatacenter(instance, vm);

                    if( dc == null ) {
                        throw new CloudException("Could not identify a deployment host.");
                    }
                    HostSystem host = getBestHost(dc);
                    
                    if( host == null ) {
                        throw new CloudException("Could not identify a deployment host.");
                    }
                    vm.powerOnVM_Task(host);
                }
                catch( TaskInProgress e ) {
                    throw new CloudException(e);
                }
                catch( InvalidState e ) {
                    throw new CloudException(e);
                }
                catch( RuntimeFault e ) {
                    throw new InternalException(e);
                }
                catch( RemoteException e ) {
                    throw new CloudException(e);
                }
            }
        }
        finally {
            if( instance != null ) {
                instance.getServerConnection().logout();
            }
        } 
    }

    VirtualMachine clone(ServiceInstance service, VirtualMachine vm, String name,  boolean asTemplate, boolean powerOn) throws InternalException, CloudException {
        try {
            name = validateName(name);
            
            Datacenter dc = ((Dc)provider.getDataCenterServices()).getVmwareDatacenter(service, getDataCenter(vm));

            Folder vmFolder = dc.getVmFolder();
            
            VirtualMachineConfigSpec config = new VirtualMachineConfigSpec();
            String sizeId = getSizeId(vm.getConfig().getHardware());
            String[] sizeInfo = sizeId.split(":");
            int cpuCount = Integer.parseInt(sizeInfo[0]);
            long memory = Long.parseLong(sizeInfo[1]);
            
            config.setName(name);
            config.setAnnotation(vm.getConfig().getAnnotation());
            config.setMemoryMB(memory);
            config.setNumCPUs(cpuCount);

            VirtualMachineCloneSpec spec = new VirtualMachineCloneSpec();
            VirtualMachineRelocateSpec location = new VirtualMachineRelocateSpec();
            ResourcePool pool = (ResourcePool) new InventoryNavigator(dc).searchManagedEntities("ResourcePool")[0];
            
            location.setHost(getBestHost(dc).getConfig().getHost());
            location.setPool(pool.getConfig().getEntity());
            spec.setLocation(location);
            spec.setPowerOn(false);
            spec.setTemplate(asTemplate);
            spec.setConfig(config);

            Task task = vm.cloneVM_Task(vmFolder, name, spec);                
            String status = task.waitForTask();
            
            if( status.equals(Task.SUCCESS) ) {
                return (VirtualMachine)(new InventoryNavigator(vmFolder).searchManagedEntity("VirtualMachine", name));
            }
            else {
                throw new CloudException("Failed to create VM: " + task.getTaskInfo().getError().getLocalizedMessage());
            }
        }
        catch( InvalidProperty e ) {
            throw new CloudException(e);
        }
        catch( RuntimeFault e ) {
            throw new InternalException(e);
        }
        catch( RemoteException e ) {
            throw new CloudException(e);
        }
        catch( InterruptedException e ) {
            throw new InternalException(e);
        }
    }
    
    @Override
    public Server clone(String serverId, String intoDcId, String name, String description, boolean powerOn, String ... firewallIds) throws InternalException, CloudException {
        ServiceInstance service = provider.getServiceInstance();
        
        try {
            VirtualMachine vm = getVirtualMachine(service, serverId);
            
            if( vm != null ) {
                return toServer(clone(service, vm, name, false, powerOn), description);
            }
            throw new CloudException("No virtual machine " + serverId + ".");
        }
        finally {
            service.getServerConnection().logout();
        }
    }

    private Random random = new Random();
    
    @Override
    public Server define(String templateId, ServerSize size, String dataCenterId, String name, boolean withAnalytics, String ... firewalls) throws InternalException, CloudException {
        ServiceInstance service = provider.getServiceInstance();

        try {
            try {
                VirtualMachine template = getTemplate(service, templateId);
            
                name = validateName(name);
                
                Datacenter dc;
                
                if( dataCenterId == null ) {
                    for( DataCenter dsdc : provider.getDataCenterServices().listDataCenters(provider.getContext().getRegionId()) ) {
                        dataCenterId = dsdc.getProviderDataCenterId();
                        if( random.nextInt()%3 == 0 ) {
                            break;
                        }
                    }
                }
                dc = ((Dc)provider.getDataCenterServices()).getVmwareDatacenter(service, dataCenterId);

                Folder vmFolder = dc.getVmFolder();
                
                VirtualMachineConfigSpec config = new VirtualMachineConfigSpec();
                String[] vmInfo = size.getSizeId().split(":");
                int cpuCount = Integer.parseInt(vmInfo[0]);
                long memory = Long.parseLong(vmInfo[1]);
                
                config.setName(name);
                config.setAnnotation(templateId);
                config.setMemoryMB(memory);
                config.setNumCPUs(cpuCount);

                VirtualMachineCloneSpec spec = new VirtualMachineCloneSpec();
                VirtualMachineRelocateSpec location = new VirtualMachineRelocateSpec();
                ResourcePool pool = (ResourcePool) new InventoryNavigator(dc).searchManagedEntities("ResourcePool")[0];
                
                location.setHost(getBestHost(dc).getConfig().getHost());
                location.setPool(pool.getConfig().getEntity());
                spec.setLocation(location);
                spec.setPowerOn(false);
                spec.setTemplate(false);
                spec.setConfig(config);

                Task task = template.cloneVM_Task(vmFolder, name, spec);                
                String status = task.waitForTask();
                
                if( status.equals(Task.SUCCESS) ) {
                    for( Server server : list() ) {
                        if( server.getName().equals(name) ) {
                            return server;
                        }
                    }
                    throw new CloudException("Unable to identify newly created server.");
                }
                else {
                    throw new CloudException("Failed to create VM: " + task.getTaskInfo().getError().getLocalizedMessage());
                }
            }
            catch( InvalidProperty e ) {
                throw new CloudException(e);
            }
            catch( RuntimeFault e ) {
                throw new InternalException(e);
            }
            catch( RemoteException e ) {
                throw new CloudException(e);
            }
            catch( InterruptedException e ) {
                throw new InternalException(e);
            }
        }
        finally {
            service.getServerConnection().logout();
        }
    }

    Architecture getArchitecture(VirtualMachineGuestOsIdentifier os) {
        if( os.name().contains("64") ) {
            return Architecture.I64;
        }
        else {
            return Architecture.I32;
        }
    }
    
    HostSystem getBestHost(Datacenter forDatacenter) throws InvalidProperty, RuntimeFault, RemoteException {
        HostSystem ohWell = null;
        
        for( ManagedEntity me : forDatacenter.getHostFolder().getChildEntity() ) {
            ClusterComputeResource cluster = (ClusterComputeResource)me;

            for( HostSystem host : cluster.getHosts() ) {
                if( host.getConfigStatus().equals(ManagedEntityStatus.green) ) {
                    return host;
                }
                if( ohWell == null || host.getConfigStatus().equals(ManagedEntityStatus.yellow) ) {
                    ohWell = host;
                }
            }
        }     
        return ohWell;
    }
    
    @Override
    public String getConsoleOutput(String serverId, long sinze) throws InternalException, CloudException {
        return "";
    }

    private String getDataCenter(VirtualMachine vm) throws InternalException, CloudException {
        ServiceInstance instance = provider.getServiceInstance();
        
        try {
            ManagedEntity parent = vm.getParent();
            
            while( parent != null ) {
                if( parent instanceof Datacenter ) {
                    return ((Datacenter)parent).getName();
                }
                parent = parent.getParent();
            }
            return null;
        }
        finally {
            if( instance != null ) {
                instance.getServerConnection().logout();
            }
        } 
    }
    
    @Override
    public Collection<String> getFirewalls(String serverId) throws InternalException, CloudException {
        return new ArrayList<String>();
    }

    private HostSystem getHost(VirtualMachine vm) throws InternalException, CloudException {
        ServiceInstance instance = provider.getServiceInstance();
        
        try {
            ManagedEntity parent = vm.getParent();
            
            while( parent != null ) {
                if( parent instanceof HostSystem ) {
                    return ((HostSystem)parent);
                }
                parent = parent.getParent();
            }
            return null;
        }
        finally {
            if( instance != null ) {
                instance.getServerConnection().logout();
            }
        } 
    }
    
    private VirtualMachine getTemplate(ServiceInstance service, String templateId) throws InvalidProperty, RuntimeFault, RemoteException {
        Folder rootFolder = service.getRootFolder();
        ManagedEntity[] mes;
        
        mes = new InventoryNavigator(rootFolder).searchManagedEntities("VirtualMachine");
        if( mes != null && mes.length > 0 ) {
            for( ManagedEntity entity : mes ) {
                VirtualMachine template = (VirtualMachine)entity;
                
                if( template != null ) {
                    VirtualMachineConfigInfo vminfo = template.getConfig();
                    
                    if( vminfo.isTemplate() && vminfo.getUuid().equals(templateId) ) {
                        return template;
                    }
                }
            }
        }
        return null;
    }
    
    @Override
    public String getProviderTermForServer(Locale locale) {
        return "server";
    }

    @Override
    public Server getServer(String serverId) throws InternalException, CloudException {
        ServiceInstance instance = provider.getServiceInstance();
        
        try {
            VirtualMachine vm = getVirtualMachine(instance, serverId);
            
            if( vm == null ) {
                System.out.println("No such VM: " + serverId);
                return null;
            }
            return toServer(vm, null);
        }
        finally {
            instance.getServerConnection().logout();
        }
    }

    @Override
    public ServerPersistence getServerPersistence() throws InternalException, CloudException {
        return ServerPersistence.ALL;
    }

    @Override
    public ServerStatistics getServerStatistics(String serverId, long start, long end) throws InternalException, CloudException {
        return null;
    }

    @Override
    public Collection<ServerStatistics> getServerStatisticsForPeriod(String serverId, long start, long end) throws InternalException, CloudException {
        return new ArrayList<ServerStatistics>();
    }

    private String getSizeId(VirtualHardware hardware) {
        return (hardware.getNumCPU() + ":" + hardware.getMemoryMB());
    }
    
    @Override
    public Collection<ServerSize> getSupportedSizes(Architecture architecture) throws InternalException, CloudException {
        ArrayList<ServerSize> sizes = new ArrayList<ServerSize>();
        
        switch( architecture ) {
            case I32:
                sizes.add(new ServerSize("1:512", "1 CPU/512 MB RAM"));
                sizes.add(new ServerSize("2:1024", "2 CPU/1 GB RAM"));
                sizes.add(new ServerSize("2:2048", "2 CPU/2 GB RAM"));
                break;
            case I64:
                sizes.add(new ServerSize("1:1024", "1 CPU/1 GB RAM"));
                sizes.add(new ServerSize("2:1024", "2 CPU/1 GB RAM"));
                sizes.add(new ServerSize("2:2048", "2 CPU/2 GB RAM"));
                sizes.add(new ServerSize("4:4096", "4 CPU/4 GB RAM"));                
                sizes.add(new ServerSize("4:10240", "4 CPU/10 GB RAM")); 
                sizes.add(new ServerSize("8:20480", "8 CPU/20 GB RAM"));
                break;
        }
        return sizes;
    }

    VirtualMachine getVirtualMachine(ServiceInstance instance, String vmId) throws CloudException, InternalException { 
        Folder rootFolder = instance.getRootFolder();
        ManagedEntity[] mes;
        
        StringBuilder debug = new StringBuilder();
        
        debug.append("Looking up " + vmId + " /" );
        try {
            mes = new InventoryNavigator(rootFolder).searchManagedEntities("VirtualMachine");
            debug.append("Folder=" + mes + " / ");
        }
        catch( InvalidProperty e ) {
            throw new CloudException("No virtual machine support in cluster: " + e.getMessage());
        }
        catch( RuntimeFault e ) {
            throw new CloudException("Error in processing request to cluster: " + e.getMessage());
        }
        catch( RemoteException e ) {
            throw new CloudException("Error in cluster processing request: " + e.getMessage());
        }
        
        if( mes != null && mes.length > 0 ) {
            for( ManagedEntity entity : mes ) {
                VirtualMachine vm = (VirtualMachine)entity;

                debug.append("VM=" + vm + " / ");
                if( vm != null ) {
                    debug.append(vm.getName() + ": " + vm.getConfig().getInstanceUuid() + " / ");
                }
                if( vm != null && vm.getConfig().getInstanceUuid().equals(vmId) ) {
                    return vm;
                }
            }
        }
        System.out.println("DEBUG=" + debug.toString());
        return null;
    }
    
    private Datacenter getVmwareDatacenter(ServiceInstance service, VirtualMachine vm) throws CloudException {
        ManagedEntity parent = vm.getParent();
        
        while( parent != null ) {
            if( parent instanceof Datacenter ) {
                return ((Datacenter)parent);
            }
            parent = parent.getParent();
        }
        return null;
    }
    
    @Override
    public boolean isSupportsCloning() {
        return true;
    }

    @Override
    public Server launch(String templateId, ServerSize size, String dataCenterId, String serverName, String withKey, String inVlanId, boolean withMonitoring, String... firewalls) throws InternalException, CloudException {
        Server server = define(templateId, size, dataCenterId, serverName, withMonitoring, firewalls);
        
        boot(server.getProviderServerId());
        return server;
    }

    @Override
    public Collection<Server> list() throws InternalException, CloudException {
        ServiceInstance instance = provider.getServiceInstance();
        
        try {
            ArrayList<Server> servers = new ArrayList<Server>();
            Folder rootFolder = instance.getRootFolder();
            ManagedEntity[] mes;
            
            try {
                mes = new InventoryNavigator(rootFolder).searchManagedEntities("VirtualMachine");
            }
            catch( InvalidProperty e ) {
                throw new CloudException("No virtual machine support in cluster: " + e.getMessage());
            }
            catch( RuntimeFault e ) {
                throw new CloudException("Error in processing request to cluster: " + e.getMessage());
            }
            catch( RemoteException e ) {
                throw new CloudException("Error in cluster processing request: " + e.getMessage());
            }
            
            if( mes != null && mes.length > 0 ) {
                for( ManagedEntity entity : mes ) {
                    Server server = toServer((VirtualMachine)entity, null);
                
                    if( server != null ) {
                        servers.add(server);
                    }
                }
            }
            return servers;
        }
        finally {
            if( instance != null ) {
                instance.getServerConnection().logout();
            }
        }
    }

    @Override
    public void monitor(String serverId) throws InternalException, CloudException {
        // NO-OP
    }

    @Override
    public void pause(String serverId) throws InternalException, CloudException {
        ServiceInstance instance = provider.getServiceInstance();
        
        try {
            VirtualMachine vm = getVirtualMachine(instance, serverId);
            
            if( vm != null ) {
                try {
                    vm.powerOffVM_Task();
                }
                catch( TaskInProgress e ) {
                    throw new CloudException(e);
                }
                catch( InvalidState e ) {
                    throw new CloudException(e);
                }
                catch( RuntimeFault e ) {
                    throw new InternalException(e);
                }
                catch( RemoteException e ) {
                    throw new CloudException(e);
                }
            }
        }
        finally {
            if( instance != null) {
                instance.getServerConnection().logout();
            }
        }
    }

    @Override
    public void reboot(String serverId) throws CloudException, InternalException {
        final String id = serverId;
        
        Thread t = new Thread() {
            public void run() {
                powerOnAndOff(id);
            }
        };
        
        t.setName("Reboot " + serverId);
        t.setDaemon(true);
        t.start();
    }
    
    private void powerOnAndOff(String serverId) {
        try {
            ServiceInstance service = provider.getServiceInstance();
            
            try {
                VirtualMachine vm = getVirtualMachine(service, serverId);
                HostSystem host = getHost(vm);

                if( vm != null ) {
                    Task task = vm.powerOffVM_Task();
                    String status = task.waitForTask();
    
                    if( !status.equals(Task.SUCCESS) ) {
                        System.err.println("Reboot failed: " + status);
                    }
                    else {
                        try { Thread.sleep(15000L); }
                        catch( InterruptedException e ) { }
                        vm = getVirtualMachine(service, serverId);
                        vm.powerOnVM_Task(host);
                    }
                }
            }
            finally {
                service.getServerConnection().logout();
            }
        }
        catch( Throwable t ) {
            t.printStackTrace();
        }
    }

    @Override
    public void stop(String serverId) throws InternalException, CloudException {
        final String id = serverId;
        
        Thread t = new Thread() {
            public void run() {
                terminate(id);
            }
        };
        
        t.setName("Reboot " + serverId);
        t.setDaemon(true);
        t.start();
    }

    private void terminate(String serverId) {
        try {
            ServiceInstance service = provider.getServiceInstance();
            
            try {
                VirtualMachine vm = getVirtualMachine(service, serverId);

                if( vm != null ) {
                    Task task = vm.powerOffVM_Task();
                    String status = task.waitForTask();
    
                    if( !status.equals(Task.SUCCESS) ) {
                        System.err.println("Termination failed: " + status);
                    }
                    else {
                        try { Thread.sleep(15000L); }
                        catch( InterruptedException e ) { }
                        vm = getVirtualMachine(service, serverId);
                        vm.destroy_Task();
                    }
                }
            }
            finally {
                service.getServerConnection().logout();
            }
        }
        catch( Throwable t ) {
            t.printStackTrace();
        }
    }
    
    static private HashMap<String,Integer> randomIps = new HashMap<String,Integer>();
    
    static private int ip = 1;
    
    private Server toServer(VirtualMachine vm, String description) throws InternalException, CloudException {
        if( vm != null ) {
            VirtualMachineConfigInfo vminfo = vm.getConfig();
            
            if( vminfo.isTemplate() ) {
                return null;
            }
            VirtualMachineGuestOsIdentifier os = VirtualMachineGuestOsIdentifier.valueOf(vminfo.getGuestId());
            Server server = new Server();
            Integer myIp = randomIps.get(vminfo.getInstanceUuid());
            
            if( myIp == null ) {
                myIp = ip++;
                randomIps.put(vminfo.getInstanceUuid(), myIp);
            }
            server.setPrivateIpAddresses("10.200.0." + myIp);
            server.setName(vm.getName());
            server.setPlatform(Platform.guess(vminfo.getGuestFullName()));
            server.setProviderServerId(vm.getConfig().getInstanceUuid());
            server.setPersistent(true);
            server.setArchitecture(getArchitecture(os));
            if( description != null ) {
                server.setDescription(description);
            }
            else {
                server.setDescription(vm.getName() + " (" + vminfo.getGuestFullName() + ")");
            }
            server.setSize(getSizeId(vminfo.getHardware()));
            String imageId = vminfo.getAnnotation();
            ServerImage img = null;
            

            if( imageId != null ) {
                img = provider.getImageServices().getImage(imageId);
            }
            if( img != null ) {
                server.setImageId(vminfo.getAnnotation());
            }
            else {  
                server.setImageId(provider.getContext().getAccountNumber() + "-unknown");
            }
            server.setRegionId(provider.getContext().getRegionId());
            server.setDataCenterId(getDataCenter(vm));

            GuestInfo guestInfo = vm.getGuest();
            
            if( guestInfo != null ) {
                String ipAddress = guestInfo.getIpAddress();
                
                if( ipAddress != null ) {
                    server.setPrivateIpAddresses(guestInfo.getIpAddress());
                    server.setPrivateDnsAddress(guestInfo.getIpAddress());
                }
            }

            VirtualMachineRuntimeInfo runtime = vm.getRuntime();
            
            if( runtime != null ) {
                VirtualMachinePowerState state = runtime.getPowerState();
                
                if( server.getCurrentState() == null ) {
                    switch( state ) {
                        case poweredOff: case suspended:
                            server.setCurrentState(ServerState.PAUSED);
                            break;
                        case poweredOn:
                            boolean pending = false;
                            
                            /*
                            for( Task task : vm.getRecentTasks() ) {
                                // TODO: review task
                                // if power on or power off, set pending = true
                            }
                            */
                            server.setCurrentState(pending ? ServerState.PENDING : ServerState.RUNNING);
                            break;
                    }
                }
                Calendar suspend = runtime.getSuspendTime();                
                Calendar time = runtime.getBootTime();

                if( suspend == null || suspend.getTimeInMillis() < 1L ) {
                    server.setStopTime(0L);
                }
                else {
                    server.setStopTime(suspend.getTimeInMillis());
                    server.setCreateTime(server.getStopTime());
                }
                if( time == null || time.getTimeInMillis() < 1L ) {
                    server.setStartTime(0L);
                }
                else {
                    server.setStartTime(time.getTimeInMillis());
                    server.setCreateTime(server.getStartTime());
                }
            }
            return server;
        }
        return null;
    }
    
    @Override
    public void unmonitor(String serverId) throws InternalException, CloudException {
        // NO-OP
    }

    private String validateName(String name) {
        name = name.toLowerCase().replaceAll("_", "-");
        if( name.length() <= 30 ) {
            return name;
        }
        return name.substring(0, 30);
    }
}
